---
elm:
  dependencies:
    avh4/elm-fifo: latest
    elm/browser: latest
    elm/core: latest
    elm/html: latest
    elm/json: latest
    elm/parser: latest
    elm/regex: latest
    elm/time: latest
    elm/url: latest
    elm/virtual-dom: latest
    elm-community/graph: latest
    elm-community/intdict: latest
    elm-community/list-extra: latest
    elm-community/maybe-extra: latest
    elm-community/string-extra: latest

  source-directories:
    - ../../src
---

# Félix y el TX-150

> Fuente: Ejercios de Lógica Computacional
> (https://www.cs.us.es/~fsancho/?p=logica-informatica-2020-21)

Félix tiene en su garaje un taller que utiliza para la reparación de máquinas. Un buen amigo suyo va a hacerle una visita y obtiene la siguiente información:

- Todos los ordenadores son máquinas.

- El TX-150 es un ordenador

- Para cualquier máquina, Félix sabe sólo cómo estropearla o sólo cómo arreglarla.

- Toda máquina puede ser arreglada por alguien (incluso por otras máquinas), pero una máquina no puede arreglarse por sí misma.

- Las cosas sólo desesperan a quienes no son capaces de arreglarlas.

- Félix está deseperado con el TX-150.

Entonces, ¿puede arreglar Félix el TX-150?

## Solución

Formalicemos el problema en el ámbito de la Lógica de Primer Orden. Para ello consideramos:

- Los símbolos de predicado: $O(x)$: "_$x$ es un ordenador_", $M(x)$: "_$x$ es una máquina_", $A(x,y)$: "_$x$ puede arreglar $y$_", $E(x,y)$: "_$x$ puede estropear $y$_" y $D(x, y)$: "_$x$ desespera a $y$_"

- Y los símbolos de constante $a$ y $b$ para representar al TX-150 y a Félix, respectivamente.

Antes de nada, vamos a importar el módulo de Sintaxis y Semántica en Primer Orden y algunos otros paquetes que nos harán falta para poder trabajar con las fórmulas.

Entonces las fórmulas asociadas a cada uno de los asertos corresponden a:

```elm {l id="imports"}

import LogicUS.FOL.SintaxSemantics exposing (..)
import LogicUS.PL.SintaxSemantics exposing (..)
import LogicUS.FOL.NormalForms exposing (..)
import LogicUS.FOL.Herbrand exposing (..)
import LogicUS.PL.Clauses exposing (..)
import LogicUS.PL.Resolution exposing (..)
import Set exposing (Set)
import Dict exposing (Dict)

ffolRead : String -> FormulaFOL
ffolRead s = ffolReadExtraction <| ffolReadFromString s
```

```elm {l m=("$ " ++ (sfolToMathString [psi1, psi2, psi3, psi4, psi5, psi6]) ++ " $") context="1"  follows="imports"}
psi1 : FormulaFOL
psi1 = ffolRead "!A[x](O(x) -> M(x))"

psi2 : FormulaFOL
psi2 = ffolRead "O(*a)"

psi3 : FormulaFOL
psi3 = ffolRead "!A[x](M(x) -> (A(*b, x) <-> ¬E(*b,x)))"

psi4 : FormulaFOL
psi4 = ffolRead "!A[x](M(x) -> (!E[y] A(y, x) & ¬A(x,x)))"

psi5 : FormulaFOL
psi5 = ffolRead "!A[x]!A[y](D(x,y) <-> ¬A(y,x))"

psi6 : FormulaFOL
psi6 = ffolRead "D(*a,*b)"
```

Ahora veámos si es posible que Félix pueda arreglar el TX-150. Esto es si el conjunto de las fórmulas anteriores junto con $A(b,a)$ es consistente.

```elm {l m=("$ " ++ (ffolToMathString psi7) ++ " $"), context="1"}
psi7 : FormulaFOL
psi7 = ffolRead "A(*b, *a)"
```

Probémos que no utilizando la extensión de Herbrand. Para ello, hemos primero de Skolemizar las fórmulas, lo cuál requiere, primero pasar las fórmulas a forma Prenex y después aplicar las adecuadas sustituciones de Skolem. Veámoslo en LogicUS.

Todas las fórmulas, excepto $\psi_4$ están ya en forma Prenex, de hecho están casi en forma de Skolem (ya que no contienen cuantificadores existenciales). Por lo que pasemos dicha fórmula también a forma Prenex. 

```elm {l m=("$ " ++ (ffolToMathString psi4p) ++ " $") context="1"}
psi4p : FormulaFOL
psi4p = ffolToPrenex psi4
```
Si queremos mostrar el proceso de construcción de la fórmula a través del árbol de formación:

```elm {l context="1"}
psi4pg : String
psi4pg = prenexGraphToDOT <| (\ (_, _, gr) -> gr)  <| ffolToPrenex2 psi4
```
<center>

```dot

digraph G {
 rankdir=TB
 graph []
 node [shape=box, color=white, fontcolor=black]
 edge [dir=none, color=blue, fontcolor=blue]

 4 -> 5
 5 -> 6
 5 -> 7
 6 -> 11
 7 -> 8
 7 -> 9
 8 -> 10 [xlabel=<∃y  >]
 9 -> 10
 10 -> 11 [label=<∃y>]
 11 -> 12

 4 [label=<∀x(M(x)→(∃yA(y,x)∧¬A(x,x)))>]
 5 [label=<(M(x)→(∃yA(y,x)∧¬A(x,x)))>]
 6 [label=<M(x)>]
 7 [label=<(∃yA(y,x)∧¬A(x,x))>]
 8 [label=<∃yA(y,x)>]
 9 [label=<¬A(x,x)>]
 10 [label=<∃y(A(y,x)∧¬A(x,x))>]
 11 [label=<∃y(M(x)→(A(y,x)∧¬A(x,x)))>]
 12 [label=<∀x∃y(M(x)→(A(y,x)∧¬A(x,x)))>]
}

```
</center>

Una vez tenemos las fórmula en la forma Prenex podemos pasar a forma de Skolem sin más que aplicar las sustituciones de Skolem pertinentes.

```elm {l m=("$ " ++ (ffolToMathString psi4sk) ++ " $") context="1"}
psi4sk : FormulaFOL
psi4sk =  ffolToSkolem psi4p
```

Podemos obtener también la sustitución de skolem realizada:


```elm {l m=("$ " ++ (substitutionToMathString psi4sk_sub) ++ " $") context="1"}
psi4sk_sub : Substitution
psi4sk_sub =  getSkolemSubs <| (Tuple.first<<extractHeaderCuantifiers) psi4p
```

Mostrado con detalle cómo realizar el paso a forma de Skolem de una fórmula vamos a pasar todo el conjunto a forma de Skolem. 

```elm {l m=("$ " ++ (sfolToMathString fs) ++ " $") context="1"}
fs : SetFOL
fs = List.map (ffolToSkolem) [psi1, psi2, psi3, psi4, psi5, psi6, psi7]
```

Ahora hemos de verificar la insatisfactibilidad del conjunto a través de la extensión de Herbrand. Nótese que la extensión asociada al problema es ifinita ya que se posee un símbolo de función (no constante). De forma que para establecer la insatisfactibilidad podemos utilizar el siguiente algoritmo:

> ##### <u>Algoritmo</u>
> + **Input**: $\Sigma$: Un conjunto de fórmulas de un Lenguaje de Primer Orden.
> + **Output**: La EH de orden mínimo que es inconsistente si el conjunto es inconsistente. En otro caso $\uparrow$
> + **Procedimiento**:
>   (1) Hacer $i \gets 1$
>   (2) Si EH(\Sigma) de orden $k$ es inconsiste entonces PARAR y DEVOLVER $i$
>   (3) Si no, hacer $i \gets i + 1$ y volver a (2)


Programemos dicho algoritmo en formato recursivo y estableciendo un orden máximo.


```elm {l context="1"}
insatConEH : SetFOL -> Int -> (Int, ClausePLSet)
insatConEH s maxO = 
    insatConEHAux s maxO 0

insatConEHAux : SetFOL -> Int -> Int ->  (Int, ClausePLSet)
insatConEHAux s maxO i =
        let
            eh_i = splToClauses <| Maybe.withDefault [] <| sfolHerbrandExtension s i
        in
        if i > maxO then
            (i, [])
        else if (Tuple.first << csplSCFResolution) eh_i then
            (i, eh_i)
        else
             insatConEHAux s maxO (i+1)

```
Veamos primero si es insatisfactible con un límite de 30. Si lo es, nos devolverá el orden mínimo que lo es, en otro caso devolverá el valor del límite más 1:

```elm {l r context="1"}
fsIsInsat30 : Int
fsIsInsat30 =  Tuple.first <| insatConEH fs 30
```

<br> En efecto, es inconsistente, y lo es la extensión de Herbrand de orden O. Veamos que la extensión de Herbrand de orden 0 corresponde a:

<!-- m=("$ " ++ (splToMathString2 fs_EH0) ++ " $") -->

```elm {l context="1"}
fs_EH0 : SetPL
fs_EH0 =  Maybe.withDefault [] <| sfolHerbrandExtension fs 0
```

$ \left\lbrace \left( O(a)_{}\rightarrow M(a)_{} \right), \, \left( O(b)_{}\rightarrow M(b)_{} \right), \, O(a)_{}, \, \left( M(a)_{}\rightarrow \left( A(b,a)_{}\leftrightarrow \neg E(b,a)_{} \right) \right), \, \left( M(b)_{}\rightarrow \left( A(b,b)_{}\leftrightarrow \neg E(b,b)_{} \right) \right), \, \left( M(a)_{}\rightarrow \left( A(ş₁(a),a)_{} \wedge \neg A(a,a)_{} \right) \right), \, \right.$ $ \left. \left( M(b)_{}\rightarrow \left( A(ş₁(b),b)_{} \wedge \neg A(b,b)_{} \right) \right), \, \left( D(a,a)_{}\leftrightarrow \neg A(a,a)_{} \right), \, \left( D(a,b)_{}\leftrightarrow \neg A(b,a)_{} \right), \, \left( D(b,a)_{}\leftrightarrow \neg A(a,b)_{} \right), \, \left( D(b,b)_{}\leftrightarrow \neg A(b,b)_{} \right), \, D(a,b)_{}, \, A(b,a)_{} \right\rbrace $

Y veamos que es inconsistente utilizando resolución proposicional: 

<!-- r=(resolutionTableauToDOT <| Tuple.second <| fs_EH0_res) -->

```elm {l r=(Tuple.first fs_EH0_res) context="1"}
fs_EH0_res : (Bool, ResolutionTableau)
fs_EH0_res = csplSCFResolution <| splToClauses fs_EH0
```



<br> Si representamos el proceso de resolución (eliminando las cláusulas irrelevantes) :


```dot
digraph G {
 rankdir=TB
 graph []
 node [shape=box, color=white, fontcolor=black]
 edge [dir=none, color=blue, fontcolor=blue]

 2 -> 17 [label=<D(a,b)>]
 3 -> 16 [label=<A(b,a)>]
 15 -> 16 [label=<¬ A(b,a)>]
 16 -> 17 [label=<¬ D(a,b)>]

 2 [label=<{D(a,b)}>]
 3 [label=<{A(b,a)}>]
 15 [label=<{¬ A(b,a),¬ D(a,b)}>]
 16 [label=<{¬ D(a,b)}>]
 17 [label=<□>]

 {rank=same; 2;3;15}
}
```